{% load wagtailcore_tags %}
{% if events %}
    <h3>Events</h3>
    <div class="upcoming-events">
      {% for title, event in events %}
      <div class="event-item clearfix">
        <div class="event-item-title"><a href="{% pageurl event %}">{{ title }}</a></div>
        {% include "joyous/includes/when_where.html" %}
      </div>
      {% endfor %}
    </div>
{% endif %}
