{% extends "joyous/joyous_base.html" %}
{% load wagtailcore_tags %}

{% block content %}
  <div class="content">
    <div class="page-heading">
      <h2>{{ page.title }}</h2>
    </div>
    <div class="content-inner">
      {{ page.intro|richtext }}

      {% block cal_options %}
      <div class="calendar-options clearfix">
        {% block events_view %}
        <div class="events-view">
          <a class="together list" href="{{ listUrl }}">Past Events</a> |
          <a class="together weekly" href="{{ weeklyUrl }}">Weekly View</a> |
          <a class="together monthly" href="{{ monthlyUrl }}">Monthly View</a>
        </div>
        {% endblock events_view %}
        <h3>Upcoming Events</h3>
      </div>
      {% endblock cal_options %}

      <div class="upcoming-events upcoming-events-detailed">
        {% for title, event in events %}
          {% include "joyous/includes/event_item.html" %}
        {% endfor %}
      </div>

      {% include "joyous/includes/events_pagination.html" %}
    </div>
    {% include "joyous/includes/event_export.html" %}
  </div>
{% endblock %}
