# Generated by Django 2.0.5 on 2018-07-09 01:51

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('joyous', '0005_auto_20180522_1158'),
    ]

    operations = [
        migrations.CreateModel(
            name='GeneralCalendarPage',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('joyous.calendarpage',),
        ),
        migrations.CreateModel(
            name='SpecificCalendarPage',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('joyous.calendarpage',),
        ),
    ]
