{% load wagtailcore_tags joyous_tags %}
<div class="event-when">
  {{ event.when }}
</div>
{% if event.postponed_from_when %}
<div class="event-postponed-from">
  Postponed from {{ event.postponed_from_when }}
</div>
{% endif %}
{% next_on event as event_next_on %}
{% if event_next_on %}
<div class="event-next-on">
  Next on {{ event_next_on }}
</div>
{% endif %}
{% if event.location %}
<div class="event-location">
  {{ event.location }}
  <a class="gmap" target="_blank" href="http://maps.google.com/?q={{ event.location }}">[map]</a>
</div>
{% endif %}
{% if event.website %}
<div class="event-website">
  <a href="{{ event.website }}" target="_blank">{{ event.website }}</a>
</div>
{% endif %}
