<style>
.event-export .input {
    box-sizing: border-box;
    border-radius: 6px;
    width: 100%;
    border: 1px solid #e6e6e6;
    padding: .9em 1.2em;
    background-color: #fafafa;
}
</style>
<li class="">
<div class="field event-export">
    <label>Download:</label>
    <div class="field-content">
        <div class="input">
            <a href="{{ self.instance.url }}?format=ical">Export ICAL</a>
            <span></span>
        </div>
    </div>
</div>
</li>

